/* GPL2 */
#define DVB_USB_LOG_PREFIX "dus01"
#include "dvb-usb.h"
#include "../frontends/va9902a.h"

#include <linux/proc_fs.h>

#define DRIVER_NAME "dus01"

MODULE_DESCRIPTION( DRIVER_NAME );
MODULE_LICENSE( "GPL" );
MODULE_VERSION( "0.1" );

static int debug = 1;
#define dbginfo(format, arg...) do{ if (debug) printk( KERN_DEBUG "dus01: " format "\n" , ## arg); }while(0)


#define DUS01_VID  0x1bc8
#define DUS01_PID  0x0001

#define COM_LEN 10

static struct dvb_usb_device *dus01_udev = 0;

static struct usb_device_id dus01_table[] =
{
    { USB_DEVICE( DUS01_VID, DUS01_PID ) },
    {}
};
MODULE_DEVICE_TABLE( usb, dus01_table );


// URB complete コールバック
// dvb_usb_data_complete() を改造
static void dus01_usb_data_complete( struct usb_data_stream *stream, u8 *buffer, size_t length )
{
    int i;

    struct dvb_usb_adapter *adap = stream->user_priv;
    if( adap->feedcount > 0 && adap->state & DVB_USB_ADAP_STATE_DVB ){

        if( length != 188 || buffer[ 0 ] != 0x47 ){
            err( "invalid TS packet" );
            return;
        }

        // 先頭除いて 0xb5 で XORかけてからリングバッファに送る
        for( i = 1; i < 188; i++ ) buffer[ i ] ^= 0xb5;
        dvb_dmx_swfilter( &adap->demux, buffer, length );
    }
}


// dvb_usb_generic_rw() を改造
static int dus01_usb_generic_rw( struct dvb_usb_device *d, int endpoint,
                           u8 *wbuf, u16 wlen, int *act_wlen,
                           u8 *rbuf,  u16 rlen, int *act_rlen,
                           int delay_ms )
{
    int ret = -ENOMEM;

    ret = mutex_lock_interruptible( &d->usb_mutex );
    if( ret ){
        err( "mutex_lock_interruptible(usb_mutex) failed" );
        return ret;
    }

    if( ! wbuf || ! act_wlen || ! wlen ) goto GENERIC_RW_ERR;

    // write
    *act_wlen = 0;
    ret = usb_bulk_msg( d->udev, usb_sndbulkpipe( d->udev, endpoint ), wbuf, wlen, act_wlen, 2000 );
    if( ret ){
        err( "usb_bulk_message(w) failed ret = %d, len = %d actlen = %d", ret, wlen, *act_wlen );
        goto GENERIC_RW_ERR;
    }

    if( rbuf && act_rlen && rlen ){

        if( delay_ms ) msleep( delay_ms );

        *act_rlen = 0;

        // read
        ret = usb_bulk_msg( d->udev, usb_rcvbulkpipe( d->udev, endpoint ), rbuf, rlen, act_rlen, 2000 );
        if (ret){
            err( "usb_bulk_message(r) failed ret = %d, len = %d actlen = %d", ret, rlen, *act_rlen );
            goto GENERIC_RW_ERR;
        }
    }

GENERIC_RW_ERR:
    mutex_unlock( &d->usb_mutex );
    return ret;
}


static int dus01_get_singnal( void )
{
    int act_wlen, act_rlen;
    u8 wbuf[ COM_LEN +1 ];
    u8 rbuf[ COM_LEN +1 ];

    memset( wbuf, 0, COM_LEN +1 );
    memset( rbuf, 0, COM_LEN +1 );

    wbuf[ 0 ] = 0x11;
    dus01_usb_generic_rw( dus01_udev, 0x01, wbuf, COM_LEN, &act_wlen, rbuf, COM_LEN, &act_rlen, 500 );

    return rbuf[ 0 ];
}


static int dus01_get_mode( void )
{
    int act_wlen, act_rlen;
    u8 wbuf[ COM_LEN +1 ];
    u8 rbuf[ COM_LEN +1 ];

    memset( wbuf, 0, COM_LEN +1 );
    memset( rbuf, 0, COM_LEN +1 );

    wbuf[ 0 ] = 0x12;
    dus01_usb_generic_rw( dus01_udev, 0x01, wbuf, COM_LEN, &act_wlen, rbuf, COM_LEN, &act_rlen, 100 );

    return rbuf[ 0 ];
}                           


static void dus01_stop( void )
{
    u8 wbuf[ COM_LEN +1 ];
    int act_wlen;

    if( dus01_get_mode() ){

        dbginfo( "stop" );

        memset( wbuf, 0, COM_LEN +1 );
        wbuf[ 0 ] = 0x21;
        dus01_usb_generic_rw( dus01_udev, 0x01, wbuf, COM_LEN, &act_wlen, NULL, 0, NULL, 0 );

        msleep( 100 );
        dbginfo( "mode = %d", dus01_get_mode() );
    }
}


// frontendからi2c経由で周波数変更と出力開始依頼
static int dus01_i2c_xfer( struct i2c_adapter *adap, struct i2c_msg msg[], int num )
{
    u8 wbuf[ COM_LEN +1 ];
    int act_wlen;
    int signal;
    int timeout = 0;
    int i;
    int len;
    struct dvb_usb_device *d = i2c_get_adapdata( adap );

    if( mutex_lock_interruptible( &d->i2c_mutex ) < 0 ){
        err( "mutex_lock_interruptible(i2c) failed" );
        return -EAGAIN;
    }

    if( num > 1 ) goto I2C_XFER_ERR;
    len = msg[ 0 ].len;
    if( len != 3 ) goto I2C_XFER_ERR;

    if( debug ){
        dbginfo( "dus01_i2c_xfer num = %d, len = %d", num , len );
        for( i = 0; i < len; ++i ) dbginfo( "[%d] = %02x", i, msg[ 0 ].buf[ i ] );
    }

    dus01_stop();

    if( ! dus01_get_mode() ){

        memset( wbuf, 0, COM_LEN +1 );
        wbuf[ 0 ] = 0x10;
        wbuf[ 1 ] = 0x04;
        wbuf[ 2 ] = 0x00;
        wbuf[ 3 ] = msg[ 0 ].buf[ 0 ];
        wbuf[ 4 ] = msg[ 0 ].buf[ 1 ];
        wbuf[ 5 ] = msg[ 0 ].buf[ 2 ];

        dus01_usb_generic_rw( dus01_udev, 0x01, wbuf, COM_LEN, &act_wlen, NULL, 0, NULL, 0 );

        timeout = 0;
        signal = 0;
        while( timeout < 50 && ! signal ){
            signal = dus01_get_singnal();
            dbginfo( "signal = %d, timeout = %d", signal, timeout );
            timeout++;
        }

        if( signal ){
            dbginfo( "start" );
            memset( wbuf, 0, COM_LEN +1 );
            wbuf[ 0 ] = 0x20;
            dus01_usb_generic_rw( dus01_udev, 0x01, wbuf, COM_LEN, &act_wlen, NULL, 0, NULL, 0 );
        }
        else err( "timeout" );
    }

I2C_XFER_ERR:

    mutex_unlock( &d->i2c_mutex );
    return num;
}

static u32 dus01_i2c_func( struct i2c_adapter *adap )
{
    return I2C_FUNC_I2C;
}

static struct i2c_algorithm dus01_i2c_algo =
{
    .master_xfer = dus01_i2c_xfer,
    .functionality = dus01_i2c_func,
};


static int dus01_pid_filter( struct dvb_usb_adapter *adap, int index, u16 pid, int onoff )
{
    dbginfo( "dus01_pid_filter index = %d, pid = %d, onoff = %d", index, pid, onoff );
    return 0 ;
}


static int dus01_streaming_ctrl( struct dvb_usb_adapter *adap, int onoff )
{
    dbginfo( "dus01_streaming_ctrl, onoff = %d", onoff );

    if( ! onoff ) dus01_stop();

    return 0;
}


// *cold = 0 を入れないと dvb_usb_find_device() が失敗する
static int dus01_identify_state(struct usb_device *udev,
                                struct dvb_usb_device_properties *prop,
                                struct dvb_usb_device_description **desc,
                                int *cold )
{
    *cold = 0;
    dbginfo( "dus01_identify_state cold = %d\n", *cold );
    return 0;
}


static int dus01_frontend_attach( struct dvb_usb_adapter *adap )
{
    dbginfo( "dus01_frontend_attach" );

    if( ( adap->fe = dvb_attach( va9902a_attach, &adap->dev->i2c_adap ) ) == NULL ){
        return -EIO;
    }

    return 0;
}


static struct dvb_usb_device_properties dus01_properties =
{
    .caps = DVB_USB_IS_AN_I2C_ADAPTER,
    .usb_ctrl = DEVICE_SPECIFIC,
    .identify_state = dus01_identify_state,
    .i2c_algo = &dus01_i2c_algo,

    .generic_bulk_ctrl_endpoint = 0x01,

    .num_adapters = 1,
    .adapter =
    {
        {
            .frontend_attach = dus01_frontend_attach,

            .caps = DVB_USB_ADAP_HAS_PID_FILTER | DVB_USB_ADAP_NEED_PID_FILTERING, // 適当

            .pid_filter_count = 16, // 適当
            .pid_filter = dus01_pid_filter,

            .streaming_ctrl = dus01_streaming_ctrl,

            .stream = {
                .type = USB_ISOC,
                .count = 10, // URBの個数 (適当)
                .endpoint = 0x83,
                .u = {
                    .isoc = {
                        .framesperurb = 4, // snoopypro の NumberOfPacketsから
                        .framesize = 188,
                        .interval = 1,  // パケット見ながら調整
                    }
                }
            },
        }
    },

    .num_device_descs = 1,
    .devices =
    {
        {
            .name = "DUS01 USB1.1",
            .warm_ids = { &dus01_table[ 0 ], NULL },
        },
        { NULL },
    }
};


static int dus01_probe(struct usb_interface *intf,
			const struct usb_device_id *id)
{
    int ret = -1;
    dbginfo( "dus01_probe" );

    ret = dvb_usb_device_init( intf, &dus01_properties, THIS_MODULE, &dus01_udev );
    if( ret ){
        err( "dvb_usb_device_init failed" );
        dus01_udev = NULL;
    }

    dus01_udev->adapter->stream.complete = dus01_usb_data_complete;

    return ret;
}


static struct usb_driver dus01_driver = {
    .name = DRIVER_NAME,
    .probe = dus01_probe,
    .disconnect = dvb_usb_device_exit,
    .id_table = dus01_table,
};


static int __init dus01_module_init( void )
{
    int ret;

    dbginfo( "dus01_module_init" );

    ret = usb_register( &dus01_driver );
    if( ret ){
        err( "usb_register failed ret = %d", ret );
        return ret;
    }

    return 0;
}


static void __exit dus01_module_exit( void )
{
    dbginfo( "dus01_module_exit" );

    usb_deregister( &dus01_driver );
}


module_init( dus01_module_init );
module_exit( dus01_module_exit );

